/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.filestore;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.nrg.xnat.tags.InvalidTagLabelException;
import org.nrg.xnat.tags.TagLabel;


/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 *
 */
public abstract class AbstractFileStore implements FileStore {
    private static final String CONTENT_TYPE_DEFAULT = "application/octet-stream";
    private static final URI[] uriArray = new URI[0];
    private static final String SQL_GET_LOCAL_FILE = "SELECT r.resource, m.mime_media_type FROM resources r, mime_media_types m WHERE r.uri = ? AND m.mime_media_type_id = r.mime_media_type_id";
    private static final String SQL_DELETE_METADATA = "DELETE FROM resource_tags WHERE resource_id IN (SELECT resource_id FROM resources WHERE uri = ?)";
    private static final String SQL_DELETE_RESOURCE = "DELETE FROM resources WHERE uri = ?";
    private static final String SQL_GET_RESOURCE_ID = "SELECT resource_id FROM resources WHERE uri = ?";
    private static final String SQL_GET_METADATA = "SELECT tl.tag_label, rt.value FROM tag_labels tl, resource_tags rt WHERE rt.resource_id = ? AND tl.tag_label_id = rt.tag_label_id";

    private final Logger log = Logger.getLogger(this.getClass());

    protected abstract InputStream getInputStream(final String path) throws IOException;

    protected abstract OutputStream getOutputStream(final String path) throws IOException;


    private final class FileStoreDataSource implements javax.activation.DataSource {
	private final String path;
	private final String name;
	private final String type;

	private FileStoreDataSource(final String path, final String name, final String type) {
	    this.path = path;
	    this.name = name;
	    this.type = null == type ? CONTENT_TYPE_DEFAULT : type;
	}

	public String getContentType() { return type; }

	public InputStream getInputStream() throws IOException {
	    return AbstractFileStore.this.getInputStream(path);
	}

	public String getName() { return name; }

	public OutputStream getOutputStream() throws IOException {
	    return AbstractFileStore.this.getOutputStream(path);
	}
    }


    /*
     * (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#getFile(java.sql.Connection, java.net.URI)
     */
    public final File getFile(final Connection c, final URI uri) throws SQLException,FileNotFoundException {
	final PreparedStatement getResource = c.prepareStatement(SQL_GET_LOCAL_FILE);
	try {
	    getResource.setString(1, uri.getPath());
	    final ResultSet rs = getResource.executeQuery();
	    try {
		if (!rs.next()) {
		    throw new FileNotFoundException("no resource found for " + uri);
		}
		final String rp = rs.getString(1);
		final String type = rs.getString(2);
		log.debug("retrieving " + type + " resource " + rp);
		return new File(rp);
	    } catch (SQLException e) {
		throw new FileNotFoundException("Unable to retrieve local resource path: " + e.getMessage());
	    } finally {
		try { rs.close(); } catch (SQLException ignore) {}
	    }
	} finally {
	    try { getResource.close(); } catch (SQLException ignore) {}
	}
    }


    /*
     * (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#getDataSource(java.sql.Connection, java.net.URI)
     */
    public final javax.activation.DataSource getDataSource(final Connection c, final URI uri)
    throws SQLException,FileNotFoundException {
	final String path = uri.getPath();
	final String[] pa = path.split("/");
	final String name = pa[pa.length - 1];
	final PreparedStatement getResource = c.prepareStatement(SQL_GET_LOCAL_FILE);
	try {
	    getResource.setString(1, path);
	    final ResultSet rs = getResource.executeQuery();
	    try {
		if (!rs.next()) {
		    throw new FileNotFoundException("no resource found for " + uri);
		}
		final String rp = rs.getString(1);
		final String type = rs.getString(2);
		log.debug("retrieving " + type + " resource " + rp);
		return new FileStoreDataSource(rp, name, type);
	    } catch (SQLException e) {
		throw new FileNotFoundException("Unable to retrieve local resource path: " + e.getMessage());
	    } finally {
		try { rs.close(); } catch (SQLException ignore) {}
	    }
	} finally {
	    try { getResource.close(); } catch (SQLException ignore) {}
	}
    }


    /**
     * Remove the named resource.  This is policy-dependent because we might want to remove
     * superdirectories as well.
     * @param path path to the resource.
     */
    protected abstract void deleteResource(String path);


    /*
     * (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#create(java.sql.Connection, java.net.URI)
     */
    public void create(final Connection c, final URI uri, final File f) throws SQLException,ForbiddenException {
	create(c, uri, f, (String)null);	// null means default media type
    }


    /* (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#remove(java.sql.Connection, boolean, java.net.URI[])
     */
    public void delete(final Connection c, boolean removeFile, final URI...uris)
    throws SQLException,FileNotFoundException {
	final PreparedStatement getResource = c.prepareStatement(SQL_GET_LOCAL_FILE);
	try {
	    final PreparedStatement deleteMetadata = c.prepareStatement(SQL_DELETE_METADATA);
	    try {
		final PreparedStatement deleteResource = c.prepareStatement(SQL_DELETE_RESOURCE);
		try {
		    for (final URI uri : uris) {
			final String suri = uri.toString();

			// Remove all associated metadata
			deleteMetadata.setString(1, suri);
			deleteMetadata.executeUpdate();

			// Remove underlying local resource, if required
			if (removeFile) {
			    getResource.setString(1, suri);
			    final ResultSet rs = getResource.executeQuery();
			    try {
				while (rs.next()) {
				    deleteResource(rs.getString(1));
				}
			    } catch (SQLException e) {
				// troublesome but not fatal
				log.error("unable to retrieve resource path for removal", e);
			    } finally {
				try { rs.close(); } catch (SQLException ignore) {}
			    }
			}

			// Remove resource from database
			deleteResource.setString(1, suri);
			if (0 == deleteResource.executeUpdate()) {
			    throw new FileNotFoundException(suri);
			}
		    }
		} finally {
		    try { deleteResource.close(); } catch (SQLException ignore) {}
		}
	    } finally {
		try { deleteMetadata.close(); } catch (SQLException ignore) {}
	    }
	} catch (RuntimeException e) {
	    try { getResource.close(); } catch (SQLException ignore) {}
	}
    }

    
    /* (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#remove(java.sql.Connection, boolean, java.util.Collection)
     */
    final public void delete(final Connection c, final boolean removeFile, final Collection<URI> uris)
    throws SQLException,FileNotFoundException {
	this.delete(c, removeFile, uris.toArray(uriArray));
    }


    @SuppressWarnings("unchecked")
    final public Map<TagLabel,Collection<String>> getMetadata(final Connection c, final URI uri)
    throws SQLException,FileNotFoundException {
	final int resourceID = getResourceID(c, uri);

	final PreparedStatement s = c.prepareStatement(SQL_GET_METADATA);
	// Load a MultiMap with the resource's metadata
	try {
	    s.setInt(1, resourceID);

	    final Map<TagLabel,Collection<String>> m = new HashMap<TagLabel,Collection<String>>();
	    final ResultSet rs = s.executeQuery();
	    try {
		while (rs.next()) {
		    final TagLabel tl;
		    try {
			tl = new TagLabel(rs.getString(1));
		    } catch (InvalidTagLabelException e) {
			// disaster!
			throw new RuntimeException("Invalid tag label in database", e);
		    }
		    if (!m.containsKey(tl)) {
			m.put(tl, new ArrayList<String>());
		    }
		    m.get(tl).add(rs.getString(2));
		}
	    } finally {
		try { rs.close(); } catch (SQLException ignore) {}
	    }
	    return m;
	} finally {
	    try { s.close(); } catch (SQLException ignore) {}
	}
    }


    public static final int getResourceID(final Connection c, final URI uri) throws SQLException,FileNotFoundException {
	final PreparedStatement getResourceID = c.prepareStatement(SQL_GET_RESOURCE_ID);
	try {
	    getResourceID.setString(1, uri.toString());
	    final ResultSet rs = getResourceID.executeQuery();
	    try {
		if (rs.next()) {
		    return rs.getInt(1);
		} else {
		    throw new FileNotFoundException(uri.toString());
		}
	    } finally {
		try { rs.close(); } catch (SQLException ignore) {}
	    }
	} finally {
	    try { getResourceID.close(); } catch (SQLException ignore) {}
	}
    }


    /*
     * (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#updateMetadata(java.sql.Connection, java.net.URI, org.nrg.xnat.filestore.MetadataOp[])
     */
    public void updateMetadata(final Connection c, final URI uri, final MetadataOp...ops)
    throws SQLException,FileNotFoundException {
	final int resourceID = getResourceID(c, uri);
	for (final MetadataOp op : ops) {
	    op.apply(resourceID);
	}
    }
}
