/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.search;

import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Logger;
import org.dom4j.Namespace;
import org.dom4j.QName;
import org.nrg.xnat.filestore.FileStore;
import org.nrg.xnat.tags.InvalidTagLabelException;
import org.nrg.xnat.tags.TagLabel;
import org.nrg.xnat.tags.TagLabelException;

/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 *
 */
public final class SearchServlet extends HttpServlet {
    private static final long serialVersionUID = 1L;
    private static final String JNDI_JDBC_DATASOURCE = "java:comp/env/jdbc/datasource";
    private static final String JNDI_XNAT_STORE = "java:comp/env/xnat/store";
    private static final String JNDI_XNAT_RESOURCE_CONTEXT = "java:comp/env/xnat/storeContext";

    private static final String QUERY_PARAM_NO_METADATA = "no-metadata";
    private static final String QUERY_PARAM_MAX_RESOURCES = "max-resources";

    private static final String XML_HEADER = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
    private static final String XMLNS_DECLARATION="xmlns=\"http://nrg.wustl.edu/xe\"";

    private static final Namespace NAMESPACE_XE = new Namespace(null, "http://nrg.wustl.edu/xe");
    private static final QName XML_RESPONSE_ROOT = new QName("SearchResults", NAMESPACE_XE);
    private static final QName XML_RESPONSE_ROOT_ATTR_QUERY = new QName("Query", NAMESPACE_XE);
    private static final QName XML_RESPONSE_RESOURCE = new QName("Resource", NAMESPACE_XE);
    private static final QName XML_RESPONSE_RESOURCE_URI = new QName("URI", NAMESPACE_XE);
    private static final QName XML_RESPONSE_TAG = new QName("Tag", NAMESPACE_XE);
    private static final String XML_RESPONSE_TAG_ATTR_LABEL = "Label";
    private static final QName XML_RESPONSE_TAG_VALUE = new QName("Value", NAMESPACE_XE);

    private final Logger log = Logger.getLogger(SearchServlet.class);

    private DataSource ds = null;
    private FileStore store = null;
    private String resourceContext = null;

    public void init(final ServletConfig config) throws ServletException {
	super.init(config);
	try {
	    final InitialContext ic = new InitialContext();
	    ds = (DataSource)ic.lookup(JNDI_JDBC_DATASOURCE);
	    store = (FileStore)ic.lookup(JNDI_XNAT_STORE);
	    resourceContext = ic.lookup(JNDI_XNAT_RESOURCE_CONTEXT).toString();
	} catch (NamingException e) {
	    throw new ServletException("Unable to obtain JDBC DataSource", e);
	}
    }

    private final static Map<String,Integer> defaultPorts = new HashMap<String,Integer>();
    static {
	defaultPorts.put("http", 80);
	defaultPorts.put("https", 443);
    }

    /*
     * Builds the indicated URL.   This is necessary mostly because there's a bug in
     * the constructor URL(String,String,int,String): no separator is put between the
     * (non-default) port and the file component.
     */
    private static URL getURL(final String protocol, final String host, final int port,
	    final String...components) throws MalformedURLException {
	final StringBuilder file = new StringBuilder();
	if (-1 != port &&
		defaultPorts.containsKey(protocol) &&
		defaultPorts.get(protocol).intValue() != port) {
	    final String first = getFirstNonEmpty(components);
	    if (null != first && '/' != first.charAt(0)) {
		file.append('/');
	    }
	}
	return new URL(protocol, host, port, joinComponents(file, '/', (Object[])components).toString());
    }

    
    private static String getFirstNonEmpty(final String[] strs) {
	for (int i = 0; i < strs.length; i++) {
	    if (null != strs[i] && ! "".equals(strs[i])) {
		return strs[i];
	    }
	}
	return null;
    }
    
    /*
     * Joins the given components by exactly one separator,
     * stripping any provided separators if necessary.
     * Empty or null components do not get separators:
     * given ('/', "c1", null, "c3") or ('/', "c1", "", "c3"),
     * produces "c1/c3" not "c1//c3"
     */
    private static StringBuilder joinComponents(final StringBuilder path, final char separator,
	    final Object...components) {
	if (components.length > 0) {
	    path.append(components[0]);
	    for (int component = 1; component < components.length; component++) {
		for (int i = path.length() - 1; separator == path.charAt(i); i--) {
		    path.deleteCharAt(i);
		}
		if (null == components[component]) continue;
		final String next = components[component].toString();
		for (int i = 0; i < next.length(); i++) {
		    if (separator != next.charAt(i)) {
			path.append(separator);
			path.append(next.substring(i));
			break;
		    }
		}
	    }
	}
	return path;
    }


    @SuppressWarnings("unchecked")
    protected void doGet(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	// Extract search criteria from parameters
	final Map<String,String[]> params = new LinkedHashMap<String,String[]>(req.getParameterMap());
	final Map<TagLabel,Collection<String>> setValues = new HashMap<TagLabel,Collection<String>>();
	final Collection<TagLabel> set = new ArrayList<TagLabel>();
	final Collection<TagLabel> unset = new ArrayList<TagLabel>();

	final boolean includeMetadata;
	if (params.containsKey(QUERY_PARAM_NO_METADATA)) {
	    includeMetadata = false;
	    params.remove(QUERY_PARAM_NO_METADATA);
	} else {
	    includeMetadata = true;
	}

	final Integer maxResources;
	if (params.containsKey(QUERY_PARAM_MAX_RESOURCES)) {
	    final String[] vals = params.remove(QUERY_PARAM_MAX_RESOURCES);
	    if (1 != vals.length) {
		final String message = QUERY_PARAM_MAX_RESOURCES + " option must have exactly one argument";
		log.debug(message);
		res.sendError(HttpServletResponse.SC_BAD_REQUEST, message);
		return;
	    }
	    maxResources = Integer.valueOf(vals[0]);
	} else {
	    maxResources = null;
	}

	// Sort the query parameters into value constraints and set/unset-tag constraints
	for (final Map.Entry<String,String[]> me : params.entrySet()) {
	    if (me.getKey().startsWith("?")) {
		// ?label returns all resources (matching any other constraints)
		// for which label is not set, plus all values for label in
		// resources that match the other constraints.
		try {
		    final TagLabel label = new TagLabel(me.getKey().substring(1));
		    unset.add(label);
		} catch (InvalidTagLabelException e) {
		    log.debug(e);
		    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
		    return;
		}
	    } else {
		final TagLabel label;
		try {
		    label = new TagLabel(me.getKey());
		} catch (InvalidTagLabelException e) {
		    log.debug(e);
		    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
		    return;
		}

		// Empty parameter means that the constraint is true if label is set to ANY value.
		// What follows is sort of nasty; the ServletRequest interface doc doesn't specify how empty
		// query parameters are handled, so in principle it could be done any of a number of ways:
		// a zero size array, or null or empty values.  This should be robust to all of these.
		final Collection<String> values = Arrays.asList(me.getValue());
		if (values.isEmpty() || values.contains(null) || values.contains("")) {
		    set.add(label);
		} else {
		    if (!setValues.containsKey(label)) {
			setValues.put(label, new ArrayList<String>());
		    }
		    setValues.get(label).addAll(values);
		}
	    }
	}

	// Matching values for unset tags come from resources that match set constraints
	final URL searchURL = new URL(req.getRequestURL().toString());
	final String protocol = searchURL.getProtocol();
	final String host = searchURL.getHost();
	final int port = searchURL.getPort();

	final Connection c;
	try {
	    c = ds.getConnection();
	} catch (SQLException e) {
	    log.error("Unable to contact database", e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}

	try {
	    final Search results;
	    try {
		results = Search.get(c, setValues, set, unset, maxResources);
	    } catch (TagLabelException e) {
		log.debug(e);
		res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
		return;
	    }

	    // Stream out: construct XML text directly and write to stream
	    final OutputStream out = res.getOutputStream();
	    out.write(XML_HEADER.getBytes());
	    final StringBuilder root = new StringBuilder("<");
	    root.append(XML_RESPONSE_ROOT.getName());
	    root.append(" ");
	    root.append(XMLNS_DECLARATION);
	    final String query = req.getQueryString();
	    if (null != query) {
		root.append(" ");
		root.append(XML_RESPONSE_ROOT_ATTR_QUERY.getName());
		root.append("=\"");
		root.append(StringEscapeUtils.escapeXml(req.getQueryString()));
		root.append("\"");
	    }
	    root.append(">");
	    out.write(root.toString().getBytes());

	    // Do tag values first because we have to collate those anyway
	    for (final TagLabel label : results.getTagLabels()) {
		final StringBuilder tagElem = new StringBuilder("<");
		tagElem.append(XML_RESPONSE_TAG.getName());
		tagElem.append(" ");
		tagElem.append(XML_RESPONSE_TAG_ATTR_LABEL);
		tagElem.append("=\"");
		tagElem.append(label);
		tagElem.append("\">");
		for (final String v : results.getTagValues(label)) {
		    tagElem.append("<");
		    tagElem.append(XML_RESPONSE_TAG_VALUE.getName());
		    tagElem.append(">");
		    tagElem.append(StringEscapeUtils.escapeXml(v));
		    tagElem.append("</");
		    tagElem.append(XML_RESPONSE_TAG_VALUE.getName());
		    tagElem.append(">");
		}
		tagElem.append("</");
		tagElem.append(XML_RESPONSE_TAG.getName());
		tagElem.append(">");
		out.write(tagElem.toString().getBytes());
	    }

	    for (final Iterator<URI> urii = results.resourceIterator(); urii.hasNext(); ) {
		final URI uri = urii.next();
		final StringBuilder uriElem = new StringBuilder("<");
		uriElem.append(XML_RESPONSE_RESOURCE.getName());
		uriElem.append("><");
		uriElem.append(XML_RESPONSE_RESOURCE_URI.getName());
		uriElem.append(">");
		uriElem.append(getURL(protocol, host, port, resourceContext, uri.toString()));
		uriElem.append("</");
		uriElem.append(XML_RESPONSE_RESOURCE_URI.getName());
		uriElem.append(">");

		if (includeMetadata) {
		    final Map<TagLabel,Collection<String>> m;
		    try {
			m = store.getMetadata(c, uri);
		    } catch (SQLException e) {
			// TODO: can we send an error this late in the process?
			log.error(e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
			return;
		    }

		    for (final Map.Entry<TagLabel,Collection<String>> me : m.entrySet()) {
			uriElem.append("<");
			uriElem.append(XML_RESPONSE_TAG.getName());
			uriElem.append(" ");
			uriElem.append(XML_RESPONSE_TAG_ATTR_LABEL);
			uriElem.append("=\"");
			uriElem.append(me.getKey());
			uriElem.append("\">");
			for (final String value : me.getValue()) {
			    uriElem.append("<");
			    uriElem.append(XML_RESPONSE_TAG_VALUE.getName());
			    uriElem.append(">");
			    uriElem.append(value);
			    uriElem.append("</");
			    uriElem.append(XML_RESPONSE_TAG_VALUE.getName());
			    uriElem.append(">");
			}
			uriElem.append("</");
			uriElem.append(XML_RESPONSE_TAG.getName());
			uriElem.append(">");
		    }

		}
		uriElem.append("</");
		uriElem.append(XML_RESPONSE_RESOURCE.getName());
		uriElem.append(">");
		out.write(uriElem.toString().getBytes());
	    }

	    final StringBuilder foot = new StringBuilder("</");
	    foot.append(XML_RESPONSE_ROOT.getName());
	    foot.append(">");
	    out.write(foot.toString().getBytes());
	} catch (SQLException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	} finally {
	    try { c.close(); } catch (SQLException ignore) {}
	}
    }
}
